#### Neil Shephard, 30th August, 2022.  "The properties of the sample median of an in-fill sequence of events 
# with an application to high frequency financial econometrics"

library(stabledist);

################ Tables 1 and 2

set.seed(111);

nRep = 25000; mRes = matrix(0,nRep,2); mStore = matrix(0,6,16);  mRes1 = mRes; mStore1 = mStore;

sigma=1; r=0.666666666; q = qnorm(0.75); m = mean(abs(rnorm(3000000))^r); v = var(abs(rnorm(3000000))^r);
estVar = (r^2)*(sigma^(2*r))/(16*(q^2)*(dnorm(q)^2));  ## for sample median
estVar1 = v/(m*m);  ## for power variation

for (i in (1:7)){    # different types of processes
  if (i==1){alpha = 0.0;}
  if (i==2){alpha = 0.3;}
  if (i==3){alpha = 0.6;}
  if (i==4){alpha = 0.9;}
  if (i==5){alpha = 1.2;}
  if (i==6){alpha = 1.5;}
  if (i==7){alpha = 1.8;}


 for (j in (1:6)){  # different sample sizes
  if (j==1){n=20};
  if (j==2){n=100};
  if (j==3){n=1000};
  if (j==4){n=10000};
  if (j==5){n=100000};
  if (j==6){n=1000000};

  mStore[j,c(1,16)] = c(n,sqrt(estVar));  # results for sample median
  mStore1[j,c(1,16)] = c(n,sqrt(estVar1));  # results for power variation

  for (k in (1:nRep)){  
    if (alpha <0.01){
      Y = sigma*rnorm(n); 
    }
    if (alpha >0.01){
      nscale = n^(0.5-(1.0/alpha)); 
      Y = (sigma*rnorm(n)) + (nscale*rstable(n,alpha,beta=0));
    }
    Z = (abs(Y))^r;
    mRes[k,] = c(k,sqrt(n)*(((median(Z)/(q^r)) - (sigma^r))));
    mRes1[k,]= c(k,sqrt(n)*(((mean(Z)/m) - (sigma^r))));
  }

  mStore[j,c(2*i,(2*i+1))] = c(mean(mRes[,2]),sd(mRes[,2]));
  mStore1[j,c(2*i,(2*i+1))] = c(mean(mRes1[,2]),sd(mRes1[,2]));
  
  print(c(alpha,n));
 }
  
}

print(mStore);  # table 1

print(mStore1);  # table 2


################################## spot volatility
############ Tables 3,4,5,6,7,8, Figure 1

set.seed(111);  

n = 2880; #### Figure 1


c1 = (1.0/(1.3))*exp(0.5*cumsum(rnorm(n)/sqrt(n)))*((0.3+(cos(2.5*(1:n)/n))^2));
c2 = (1.0/(1.3))*exp(0.5*cumsum(rnorm(n)/sqrt(n)))*((0.3+(cos(2.5*(1:n)/n))^2));
c3 = (1.0/(1.3))*exp(0.5*cumsum(rnorm(n)/sqrt(n)))*((0.3+(cos(2.5*(1:n)/n))^2));

pdf("volsim.pdf");
  plot((1:n)/n,c1,type="l",ylab="Vol",xlab="Time",ylim=c(0,2));
  points((1:n)/n,c2,type="l",col="red",ylim=c(0,2));
  points((1:n)/n,c3,type="l",col="blue",ylim=c(0,2));
dev.off();


#######

q = qchisq(0.5,1)^0.5; 
nRep = 100000; mRes = matrix(0,nRep,3); mStore = matrix(0,6,17);
mRes1 = mRes; mStore1 = mStore;  mRes2 = mRes; mStore2 = mStore; mRes3 = mRes; mStore3 = mStore;


for (h in (1:2)){
  if (h==1){
    r=0.66666666; # Tables 3 and 4 
  }
  if (h==2){
    r = 2.0; # Table 5, appendix  
  }


set.seed(111);

estVar = (r^2)/(16*(q^2)*(dnorm(q)^2)); ## for sample median 
m = mean(abs(rnorm(3000000))^r);
mhalf = mean(abs(rnorm(3000000))^(r/2));
v = var(abs(rnorm(3000000))^r);
estVar2 = (m^2 + (2.0*m*mhalf*mhalf) - (3.0*mhalf^4))/(mhalf^4); # bipower variation
estVar1 = v/(m*m); ## for power variation
estVar4 = (0.6666666^2)/(16*(q^2)*(dnorm(q)^2)); ## for sample median and r=2/3.

for (i in (1:7)){  # different types of processes
  if (i==1){alpha = 0.0;}
  if (i==2){alpha = 0.3;}
  if (i==3){alpha = 0.6;}
  if (i==4){alpha = 0.9;}
  if (i==5){alpha = 1.2;}
  if (i==6){alpha = 1.5;}
  if (i==7){alpha = 1.8;}
  
  
  for (j in (1:5)){  # different sample sizes
    if (j==1){n=200};
    if (j==2){n=1000};
    if (j==3){n=10000};
    if (j==4){n=100000};
    if (j==5){n=1000000};
    
#    nt = ceiling(0.7*(n^0.44));
    nt = ceiling(2*(n^0.44));
    
    mStore[j,c(1,2,17)] = c(n,nt,sqrt(estVar));  # results for sample median
    mStore1[j,c(1,2,17)]= c(n,nt,sqrt(estVar1));  # results for power variation 
    mStore2[j,c(1,2,17)]= c(n,nt,sqrt(estVar2));  # results for bipower variation 
    shri = 1.0+(3.0*estVar4/nt);
    mStore3[j,c(1,2,17)]= c(n,nt,sqrt(estVar));  # results for shrunk sample median 


 for (k in (1:nRep)){
  w = cumsum(rnorm(nt)/sqrt(n));
  c= (1.0/(1.3))*exp(0.5*cumsum(rnorm(nt)/sqrt(n)))*((0.3+(cos(2.5*(1:nt)/n))^2));

  if (alpha <0.01){
    y = rnorm(nt)*c;
  }
  if (alpha >0.01){
    nscale = n^(0.5-(1.0/alpha)); 
    y = (rnorm(nt)*c) + (nscale*rstable(nt,alpha,beta=0));
  }
  z = abs(y)^r;

  est1 = median(z)/(q^r);  # usual median
  u = sqrt(nt)*(est1 - (c[1]^r));
  
  mRes[k,] = c(est1,u,u);
  
  est2 = (mean(z)/m);      # power variation
  u2 = sqrt(nt)*(est2-(c[1]^r));
  mRes1[k,] = c(est2,u2,u2);
  
  est3 = mean(sqrt(z[1:(nt-1)]*z[2:nt]))/(mhalf^2);  # bipower
  u3 = sqrt(nt)*(est3 - (c[1]^r));
  mRes2[k,] = c(est3,u3,u3);
  
  est44 = (median(abs(y)^0.666666))/(q^0.666666); # shrunk median for r=2
  shri = 1.0+(3.0*estVar4/nt);
#  est4 = (est44^3)*exp(-3.0*estVar4/nt);
  est4 = (est44^3)/shri;
  u4 = sqrt(nt)*(est4 - (c[1]^2));
  mRes3[k,] = c(est4,u4,u4);

 }
    
    mStore[j,c((2*i+1),(2*i+2))] = c(mean(mRes[,2]),sd(mRes[,2]));
    mStore1[j,c((2*i+1),(2*i+2))]= c(mean(mRes1[,2]),sd(mRes1[,2]));
    mStore2[j,c((2*i+1),(2*i+2))]= c(mean(mRes2[,2]),sd(mRes2[,2]));
    mStore3[j,c((2*i+1),(2*i+2))]= c(mean(mRes3[,2]),sd(mRes3[,2]));

#    print(c(alpha,n,nt,mean(mRes[,2]),sd(mRes[,2]),(1^r)*sqrt(estVar)));
#    print(c(alpha,n,nt,mean(mRes1[,2]),sd(mRes1[,2])));
  }
  print(alpha);
}

 if (h==1){  # main text, Tables 3,4,5
  print("main text, median, r=2/3");  print(mStore); # median
  print("main text, power, r=2/3"); print(mStore1);  # power 
   print("main text, bipower, r=2/3"); print(mStore2); # bipower
 }

 if (h==2){  # Appendix, Tables 6,7,8
  print("Appendix, median,r=2"); print(mStore);  # median
  print("Appendix, bipower,r=2"); print(mStore2);  # bipower
  print("Appendix, shrunk median,r=2"); print(mStore3);  # shrunk median
 }


}

#######
### estimating vol sequence

set.seed(111);

alpha = 0.7; n =2880; nt = ceiling(0.7*(n^0.44)); q = qchisq(0.5,1)^0.5; 
estVar4 = (0.6666666^2)/(16*(q^2)*(dnorm(q)^2)); 
mhalf = mean(abs(rnorm(3000000)));
mRes = matrix(0,n,5); 

nt = ceiling(2*(n^0.44));
shri = mean(abs(1.0+sqrt(estVar4/nt)*rnorm(1000000))^1.5);

for (k in (1:3)){
  


if (alpha <0.01){
  y = rnorm(n)*c1;
}
if (alpha >0.01){
  nscale = n^(0.5-(1.0/alpha)); 
  y = (rnorm(n)*c1) + (nscale*rstable(n,alpha,beta=0));
}

for (i in (nt:n)){
  ysub = y[(i-nt+1):i];
  est44 = (median(abs(ysub)^0.666666))/(q^0.666666); # shrunk median for r=2
  volest = (est44^(3/2))/shri;
  volBPV = sqrt(mean(abs(ysub[1:(nt-1)]*ysub[2:nt]))/(mhalf^2));
  volRV = sqrt(mean(ysub[1:nt]^2));
  
  mRes[i,c(3,4,5)] = c(volest,volBPV,volRV);
}
mRes[,2] = c1;

mRes[(1:(nt-1)),3] = mRes[nt,3];
mRes[(1:(nt-1)),4] = mRes[nt,4];
mRes[(1:(nt-1)),5] = mRes[nt,5];


 pdf(paste("volseq",k,".pdf",sep=""));
  plot((nt:n)/n,mRes[(nt:n),2],type="l",col="black",lwd=5,ylim=c(0,2),ylab="Spot vol",xlab="Time");
  points((nt:n)/n,mRes[(nt:n),3],type="l",col="red",ylim=c(0,2));
 dev.off();

 pdf(paste("volseqest",k,".pdf",sep=""));
  plot((nt:n)/n,mRes[(nt:n),2],type="l",col="black",lwd=3,ylim=c(0,2),ylab="Spot vol",xlab="Time");
  points((nt:n)/n,mRes[(nt:n),3],type="l",col="red",ylim=c(0,2));
  points((nt:n)/n,mRes[(nt:n),4],type="l",col="green",ylim=c(0,2));
  points((nt:n)/n,mRes[(nt:n),5],type="l",col="blue",ylim=c(0,2));
 dev.off();
}







###### playing

nRep = 100;
n=288; nt = ceiling(1.2*sqrt(n)); r = 0.666666; q = qchisq(0.5,1)^0.5;
m = mean(abs(rnorm(3000000)));
xProcess = matrix(0,n,5);
nscale = n^(0.5-(1.0/alpha));  
alpha = 0.0;

mRes = matrix(0,nRep,6);

for (k in (1:nRep)){
  

w = cumsum(rnorm(n)/sqrt(n));
c= exp(0.5*w)*(0.1+cos(2.6*(1:n)/n)^2);

if (alpha <0.01){
  y = rnorm(n)*c;
}
if (alpha >0.01){
  nscale = n^(0.5-(1.0/alpha)); 
  y = (rnorm(n)*c) + (nscale*rstable(n,alpha,beta=0));
}

nscale = n^(0.5-(1.0/alpha));

for (i in (1:n)){
  if (i<(n-nt)){
    ysub = y[i:(nt+i-1)];
    z = abs(ysub)^r;
    est1 = median(z)/(q^r);
    
    BPV = sqrt(mean(abs(ysub[1:(nt-1)])*abs(ysub[2:(nt)]))/(m^2));
    
  }
  xProcess[i,] = c(i,est1^(1.0/r),c[i],abs(y[i]),BPV);
}



RV = mean(xProcess[,4]^2);
IV = mean(xProcess[,3]^2);
medRV = mean(xProcess[,2]^2);
BPV = mean(xProcess[,5]^2);

#print(c(RV,medRV,BPV,IV,RV-IV,medRV-IV));

mRes[k,] = c(RV,medRV,BPV,IV,BPV-IV,medRV-IV);

}
mStore = c(mean(mRes[,5]),sd(mRes[,5]),mean(mRes[,6]),sd(mRes[,6]));



plot((1:n)/n,xProcess[,2],type="l",col="red",ylab="vol",xlab="t");
points((1:n)/n,xProcess[,3],type="l",col="black");
points((1:n)/n,xProcess[,5],type="l",col="orange");










